fs	   = require 'fs'
{exec} = require 'child_process'
util   = require 'util'
uglify = require 'uglify-js'
stalker = require 'stalker'
growl = require 'growl'
path = require 'path'
async = require 'async'

libDir			= path.join(path.dirname(fs.realpathSync(__filename)), '../lib')
rootDir			= process.cwd()
config = try require "#{rootDir}/dependency/kitkat_config.js" # Provides overrides for paths used.
config = config?={} # if no config file has been defined this will make it into an empty object and use default values in this sourcecode

prodSrcDir		= config.prodSrcDir?="#{rootDir}/src"		# assumed that code files could be within the project anywhere
specFileDir		= config.specFileDir?="#{rootDir}/spec"
prodDeployDir	= config.prodDeployDir?="#{rootDir}/lib"	# This is the location for compiled code so that tests may be run against them, each file is compliled here as well as a combine js
commonJsMock	= config.commonJsMock?="/mocks/commonJs.js"	# This file will contain any generic mock ups required by tests, individual mock ups needed in a single spec will be done in the spec file itself 

prodTargetCreate = config.prodTargetCreate?=false							# When true the combined compilation occurs, can be set by config file or command argument "uglify"
prodTargetFileName	 = path.basename(rootDir)								# These variables detemerine the generated combined coffee, .js and minified .js files generated
prodTargetCoffeeFile = "#{prodDeployDir}/#{prodTargetFileName}.app.coffee"		# they will be used to generate a deployable code file as well as used in the combined test of all sepcs
prodTargetJsFile	 = "#{prodDeployDir}/#{prodTargetFileName}.app.js"		# against the single .min.js file for any errors which may occur based on src files using like names in src code
prodTargetJsMinFile	 = "#{prodDeployDir}/#{prodTargetFileName}.min.js"		# both spec and code files are compiled here, spec files will be filename.spec.js and codefiles filename.js

# Variables to support the three supported Testing engines
engine = {}
engine.jasmine_node = path.join(path.dirname(fs.realpathSync(__filename)), '../node_modules/jasmine-node/bin/jasmine-node spec/')
engine.rhino = "#{libDir}/env_jasmine/bin/run_test.sh --projectPath=#{rootDir} "
engine.expresso = path.join(path.dirname(fs.realpathSync(__filename)), '../node_modules/expresso/bin/expresso spec/*.coffee -I lib')
engine.type = "expresso"
engine.execute = engine.expresso # default

prodCoffeeFiles = []
specCoffeeFiles = []
prodJsFiles = []
specJsFiles = []
intialTestComplete = false
autoSpec = false
showCoverageReport = false

prodCoffeeOpts = "--bare --output #{prodDeployDir} --compile #{prodTargetCoffeeFile}"
specCoffeeOpts = "--output #{specFileDir}"


# Verify paths exist, create if they do not, send back error if any issues occured.
checkPaths = (callback) ->
	path.exists prodDeployDir, (exists) ->
		if not exists
			exec "mkdir #{prodDeployDir}", (err, stdout, stderr) ->
				if err
					callback err
	path.exists specFileDir, (exists) ->
		if not exists
			exec "mkdir #{specFileDir}", (err, stdout, stderr) ->
				if err
					callback err
	path.exists prodSrcDir, (exists) ->
		if not exists
			exec "mkdir #{prodSrcDir}", (err, stdout, stderr) ->
				if err
					callback err
	mockDir = rootDir + path.dirname(commonJsMock)
	path.exists (rootDir + commonJsMock), (exists) ->
		if not exists
			exec "mkdir #{mockDir}", (err, stdout, stderr) ->
				if err
					callback err
				else
					exec "cp #{libDir}/commonjs.js #{mockDir}/commonjs.js", (err,stdout, stderr) ->	#Make iniitial base file
						callback err if err
	callback

# Entry point to kitkat
exports.run = ->
	# Commandline processor
	checkPaths (err) ->
		if err
			handleError err
			return
		
	optimist = require 'optimist'
	args = optimist.argv
	optimist.alias 'test', 't'
	optimist.alias 'spec', 's'
	optimist.alias 'autospec', 'a'
	optimist.alias 'engine', 'e'
	optimist.alias 'uglify', 'u'
	optimist.alias 'coverage', 'c'
	optimist.usage("Kitkat - Kontinuous Integrated Testing Koffee Application Template\n\n default usage with no arguments monitors current project for changes and runs all spec files found against all coffee files found.", {
		'test': {description: 'runs all tests against your application js files in the spec\n\t\tpath unless overridden in the dependency config file\n', required: false},
		'spec': {description: 'generates spec files based on the supplied path/file, if none\n\t\tprovided then processes all coffee files in project\n', required: false},
		'autospec': {description: 'generates new spec files and monitors all files for changes,\n\t\tautomatically runs tests on startup and as files change\n', required: false},
		'engine': {description: 'specifies the testing engine to use, by default we will use expresso\n\t\tavailable options include "expresso", "rhino", and "jasmine-node"\n', required: false} ,
		'ugilfy': {description: 'creates a combined source file then minifies it using uglify-js library', required: false},
		'coverage': {description: 'generates the code coverage report at the end of testing', required: false}
		})
	commands = []
	for key, value of args
		if key != "$0" and key != "_"
			try 
				switch key.toLowerCase()
					when "t", "test", "s", "spec", "a", "autotest"
						commands.push key
					when "e", "engine"
						engine.type = value?.toLowerCase?()
						switch engine.type
							when "expresso"
								engine.execute = engine.expresso
							when "rhino"
								engine.execute = engine.rhino
							when "jasmine-node"
								engine.execute = engine.jasmine_node
							else
								throw "Invalid engine"
					when "u", "uglify"
						prodTargetCreate = true
					when "c", "coverage"
						showCoverageReport = true
						engine.expresso = engine.expresso + " --coverage"
						engine.execute = engine.expresso if engine.type == "expresso"
					else
						throw "Invaid argument"
			catch ex
				util.log "#{ex}\n"
				optimist.showHelp() # invalid engine specificed, you do not have to specify an engine to use expresso
				return

	command = commands[0]?.toLowerCase()

	switch command
		when "t", "test"
			util.log "Initiating tests......."
			test() # args.t ?= args.test
		when "s", "spec"
			util.log "Generating spec files......"
			spec args.s ?= args.spec
		when "a", "autospec"
			util.log "Autospec and testing current project....."
			autospec()
		else
			util.log "Monitoring and testing current project....."
			main()

#########################################
# Functions follow in alphabetical order
#########################################

autospec = ->
	autoSpec = true
	main()

coffee = (options = "", file) ->
	util.log "Compiling #{file}"
	exec "coffee #{options} --compile #{file}", (err, stdout, stderr) -> 
		handleError(err) if err

combineProd = (createCombined = false, callback) ->
	#make the combined .coffee file, app.js and app.min.js files
	appContents = new Array remaining = prodCoffeeFiles.length
	util.log "Appending #{prodCoffeeFiles.length} files to #{prodTargetCoffeeFile}"
	for file, index in prodCoffeeFiles 
		coffee "--output #{prodDeployDir}", "#{file}"
		if createCombined
			fs.readFile "#{file}", 'utf8', (err, fileContents) ->
				callback err if err
				appContents[index] = fileContents
				process() if --remaining is 0
		else
			callback null if --remaining is 0

	process = ->
		fs.writeFile prodTargetCoffeeFile, appContents.join('\n\n'), 'utf8', (err) ->
			callback err if err
		exec "coffee #{prodCoffeeOpts}", (err, stdout, stderr) ->
			callback err if err
			message = "Compiled #{prodTargetJsFile}"
			util.log message
			callback null

displayNotification = (message = '') -> 
	options = {
		title: 'CoffeeScript'
		image: 'lib/CoffeeScript.png'
	}
	growl.notify message, options

gatherFiles = (filepath, extRegex, callback) ->
	filesFound = []
	file = path.normalize(filepath)
	fs.stat file, (err,stats) ->
		if err
			file += extRegex.toString().replace(/\//g,"").replace(/\\/g,"")
			fs.stat file, (err,stats) ->
				callback err, null if err 
				if not err
					filesFound.push file 
					callback null, filesFound
		else
			if stats.isDirectory()
				walk path.normalize(file + '/'), extRegex, (err,results) ->
					callback err, null if err
					if not err
						filesFound = results	
						callback null, filesFound
			else
				filesFound.push file
				callback null, filesFound

handleError = (error) -> 
	util.log error
	displayNotification error

main = ->
	stalker.watch "#{prodSrcDir}" , watchCoffeeFile, unWatchCoffeeFile
	stalker.watch "#{specFileDir}" , watchCoffeeFile, unWatchCoffeeFile
	test "/"

runTests = (callback) ->
	message = ""
	exec engine.execute, (err, stdout, stderr) ->
		util.log engine.execute
		switch engine.type
			when "rhino"
				pass = stdout.match(/Passed: \d+/)
				fail = stdout.match(/Failed: \d+/)
				total = stdout.match(/Total : \d+/)
				message = "#{pass} #{fail} #{total}"
				util.log stdout
			when "jasmine-node"
				message = "" + stdout.match(/.*failure/).toString()
				output = stdout.toString()
				output = output.replace(/at .*\n/g, "") # Strip out trace stack
				test = message.match(/\d+ test/)
				assertion = message.match(/\d+ assertion/)
				failure = message.match(/\d+ failure/);
				message = "#{test}s, #{assertion}s, #{failure}"
				util.log output
			else # "expresso"
				if err
					message = "" + stderr.match(/Failures.*/)?.toString()		# Extract failure count
					message = message.replace(/\[\d*m/g, "") 					# stripping special formatting from output
					output = stderr.toString()
					output = output.replace(/at .*\n/g, "")
					util.log output
				else
					message = "100% - " + stderr.match(/\d* tests/)?.toString() + " passed"		# Extract Test count
		if err
			options = {
				title: 'Automated Testing'
				image: "#{libDir}/error.png"
			}
		else
			options = {
				title: 'Automated Testing'
				image: "#{libDir}/success.png"
			}

		growl.notify message, options
		callback null

spec = (filepath) ->
	async.series
		gather_files: (callback) ->
			filepath = "/" if (filepath==true)
			filepath = path.join(prodSrcDir, path.normalize(filepath))
			gatherFiles filepath, /\.coffee/, (err, results) ->
				prodCoffeeFiles = results
				callback err, results
		create_specs: (callback) ->
			pending = prodCoffeeFiles.length
			for file, index in prodCoffeeFiles then do (file, index) ->
				specFile = path.join(specFileDir, path.basename(file, '.coffee') + ".spec.coffee")
				fs.stat specFile, (err,stats) ->
					if err # File does not exist
						prodFile = path.join(prodDeployDir, path.basename(file, '.coffee') + ".js")
						writeSpecFile specFile, prodFile, (err, results) ->
							callback err, null if err
				callback null, null unless --pending
	, (err, results) ->
		handleError(err) if err
		

test = ->
	async.series
		gather_files: (callback) ->
			async.parallel
				gather_specs: (callback) ->
					gatherFiles specFileDir, /\.spec\.coffee/, (err, results) ->
						specCoffeeFiles = results
						callback err, results
				gather_prod: (callback) ->
					gatherFiles prodSrcDir, /\.coffee/, (err, results) ->
						prodCoffeeFiles = results
						callback err, results
			, (err, results) ->
				callback err, results

		create_specs: (callback) ->
			if autoSpec
				pending = prodCoffeeFiles.length
				for file, index in prodCoffeeFiles then do (file, index) ->
					specFile = path.join(specFileDir, path.basename(file, '.coffee') + ".spec.coffee")
					fs.stat specFile, (err,stats) ->
						if err # File does not exist
							prodFile = path.join(prodDeployDir, path.basename(file, '.coffee') + ".js")
							writeSpecFile specFile, prodFile, (err, results) ->
								callback err, null if err
							specCoffeeFiles.push specFile
					callback null, null unless --pending
			else 
				callback null, null

		combine_prod: (callback) ->
			combineProd prodTargetCreate, (err) ->
				callback err,null

		compile_specs: (callback) ->
			pending = specCoffeeFiles.length
			for file, index in specCoffeeFiles then do (file, index) ->
				coffee specCoffeeOpts, "#{file}"
				callback  null, null unless --pending

		run_specs: (callback) ->
			runTests (err) ->
				intialTestComplete = true
				callback err, null

		uglifyIt: (callback) ->
			if not prodTargetCreate
				callback null,null
			else
				jsp = uglify.parser
				pro = uglify.uglify
				fs.readFile prodTargetJsFile, 'utf8', (err, fileContents) ->
					ast = jsp.parse fileContents  # parse code and get the initial AST
					ast = pro.ast_mangle ast # get a new AST with mangled names
					ast = pro.ast_squeeze ast # get an AST with compression optimizations
					final_code = pro.gen_code ast # compressed code here

					fs.writeFile prodTargetJsMinFile, final_code
					message = "Uglified #{prodTargetJsMinFile}"
					util.log message
				callback null, null
	, (err, results) ->
		handleError(err) if err

unWatchCoffeeFile = (err, file) ->
	if "#{file}".match(/\.coffee/)?
		util.log "#{file} removed."
		fs.unwatchFile "#{file}"
		runTests (err) ->
			return

walk = (dir, extRegex, callback) ->
	results = []
	pending = 0
	fs.readdir dir, (err, list) ->
		callback err, null if err
		pending = list.length
		for file in list then do(file) ->
			file = dir + file
			fs.stat file, (err, stat) ->
				if stat and stat.isDirectory()
					file += "/"
					if "#{file}".match(/node_modules/)?
						callback null,results unless --pending
					else
						walk file, extRegex, (err, res) ->
							results = results.concat(res)
							callback null, results unless --pending
				else
					if file.match(extRegex)
						results.push file
					callback null, results unless --pending

watchCoffeeFile = (err, file) ->
	if not "#{file}".match(/\.spec\.coffee/)? and "#{file}".match(/\.coffee/)?
		spec "/" if (autoSpec==true and intialTestComplete==true)
	if "#{file}".match(/\.coffee/)?
		util.log "Watch added for #{file}"
		fs.unwatchFile "#{file}"
		fs.watchFile "#{file}", (curr, prev) ->
			if +curr.mtime isnt +prev.mtime
				util.log "Saw change in #{file}"
				if "#{file}".match(/\.spec\.coffee/)
					coffee specCoffeeOpts, "#{file}"
				else
					coffee "--output #{prodDeployDir}", "#{file}"
				runTests (err) ->
					return
					
	
writeSpecFile = (specFile, prodFile, callback) ->
	libFile = path.basename(prodFile, path.extname(prodFile))
	switch engine.type
		when "rhino"
			specContent = ["EnvJasmine.load (EnvJasmine.projectPath + \"", "EnvJasmine.load (EnvJasmine.projectPath + \"" , "describe 'default spec suite', ->", "\tit 'needs to be modified with real test cases and thus fails until this file is modified.', ->", "\t\t(expect false).toBeTruthy()" ]
			specContent[0] = (specContent[0]) + commonJsMock + "\")"
			specContent[1] = (specContent[1]) + prodDeployDir.replace(rootDir,"") + "/" +  path.basename(prodFile) + "\")"
		when "jasmine-node"
			specContent = ["replace with libfile" , "describe 'default spec suite', ->", "\tit 'needs to be modified with real test cases and thus fails until this file is modified.', ->", "\t\t(expect false).toBeTruthy()" ]
			specContent[0] = "#{libFile} = require('../lib/#{libFile}')"
		else # expresso
			specContent = ["replace with libfile", "assert = require('assert')", "module.exports =", "\t\"You need to modify this with real test cases and thus it fails until this file is modified.\": ->", "\t\tassert.ok false, \"it will fail\"" ]
			specContent[0] = "#{libFile} = require('#{libFile}')"

	fs.writeFile specFile, specContent.join('\n'), 'utf8', (err) ->
		callback err, null if err
		util.log "Created specfile '#{specFile}'"
	callback null, null	